""" REST helpers for Muffin Framework. """

from muffin import HTTPNotFound, HTTPBadRequest, HTTPForbidden

# Package information
# ===================

__version__ = "0.0.25"
__project__ = "muffin-rest"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"


class RESTNotFound(HTTPNotFound):

    """ Custom excption class for stopping default application error handlers. """

    pass


class RESTBadRequest(HTTPBadRequest):

    """ Custom excption class for stopping default application error handlers. """

    pass


class RESTForbidden(HTTPForbidden):

    """ Custom exception class for stopping default application error handlers. """

    pass

from .api import *      # noqa
from .filters import *  # noqa
from .forms import *    # noqa
from .handlers import * # noqa

try:
    from .peewee import PWRESTHandler           # noqa
    from .peewee import PWFilter, PWLikeFilter  # noqa
except ImportError:
    pass



""" .

!!! STILL NOT IMPLEMENTED. WORK IN PROGRESS !!!

"""

import re
import muffin
import asyncio
from types import FunctionType
from aiohttp import web


PREFIX_RE = re.compile('(/|\s)')


class ApiRoute(web.Route):

    """ Support multiplie routers. """

    def __init__(self, api):
        """ Init router. """
        self.api = api
        self._name = api.prefix_name
        self._path = api.prefix
        self._method = "*"
        self.router = web.UrlDispatcher()

    def match(self, path):
        """ Check path for API. """
        if path.startswith(self.api.prefix):
            path = path[len(self.api.prefix):]
            for name, route in self.router.items():
                match_info = route.match(path)
                if match_info is not None:
                    match_info['api-name'] = name
                    return match_info
        return None

    @asyncio.coroutine
    def _handler(self, request):
        route = self.router[request.match_info['api-name']]
        response = yield from route._handler(request)
        return response

    def url(self, **kwargs):
        """ Do nothing for now. """
        pass


class Api():

    """ Support API. """

    def __init__(self, app, prefix='/api', scheme=False):
        """ Initialize the API. """
        self.app = app
        self.prefix = prefix.rstrip('/')
        self.prefix_name = PREFIX_RE.sub('-', prefix.strip('/'))
        self.handlers = {}
        self.urls = ApiRoute(self)
        self.app.router.register_route(self.urls)
        if scheme:
            path = '/'.join((self.prefix, scheme.strip('/')))
            handler = muffin.Handler.from_view(self.render_scheme, 'GET')
            app.register(path)(handler)

    def register(self, *paths, methods=None, name=None):
        """ Register handler to application. """
        if isinstance(methods, str):
            methods = [methods]

        def wrapper(handler):

            if isinstance(handler, FunctionType):
                handler = muffin.Handler.from_view(handler, *(methods or ['GET']))

            if handler.name in self.handlers:
                raise muffin.MuffinException('Handler is already registered: %s' % handler.name)

            self.handlers[handler.name] = handler

            handler.connect(
                self.app, *paths, methods=methods,
                name=name or "%s-%s" % (self.prefix_name, handler.name), router=self.urls.router)

            return handler

        # Support for @app.register(func)
        if len(paths) == 1 and callable(paths[0]):
            view = paths[0]
            paths = []
            return wrapper(view)

        return wrapper

    def render_scheme(self, request):
        """ Render API Scheme. """
        response = {}

        for name, handler in self.handlers.items():
            scheme = {'methods': handler.methods, 'desc': handler.__doc__}
            scheme.update(getattr(handler, 'scheme', lambda: {})())
            response[name] = scheme

        return response



""" Support admin filters. """

import wtforms as wtf


FILTER_PREFIX = 'mr-'


class FilterDefault:

    """ Default filters value. """

    def __str__(self):
        """ Nothing here. """
        return ""


class FilterForm(wtf.Form):

    """ Store filters for resource. """

    def process(self, collection, formdata=None, obj=None, data=None, **kwargs):
        """ Filter collection. """
        super(FilterForm, self).process(formdata, obj, data, **kwargs)
        for field in self._fields.values():
            collection = field.flt.filter(collection, self.data)
        return collection


class Filter:

    """ Implement filters. """

    field = wtf.StringField
    default = FilterDefault()
    options = {}

    def __init__(self, name, **options):
        """ Store name and mode. """
        self.name = name
        self.options = options or self.options

    def bind(self, form):
        """ Bind to filter's form. """
        field = self.field(**self.options)
        field = form._fields[self.name] = field.bind(form, self.name, prefix=form._prefix)
        field.flt = self

    def filter(self, collection, data):
        """ Load value and filter collection. """
        value = self.value(data)
        if value is self.default:
            return collection
        return self.apply(collection, value)

    def value(self, data):
        """ Get value from data. """
        value = data.get(self.name, self.default)
        return value or self.default

    def apply(self, collection, value):
        """ Filter collection. """
        return [o for o in collection if getattr(o, self.name, None) == value]


class BoolFilter(Filter):

    """ Boolean filter. """

    field = wtf.BooleanField


class IntegerFilter(Filter):

    """ Integer filter. """

    field = wtf.IntegerField


class ChoiceFilter(Filter):

    """ Boolean filter. """

    field = wtf.SelectField


def default_converter(handler, flt):
    """ Convert column name to filter. """
    if isinstance(flt, Filter):
        return flt
    return Filter(flt)



from wtforms import validators, widgets         # noqa
from wtforms.fields import *                    # noqa
from wtforms.form import Form                   # noqa
from wtforms.validators import ValidationError  # noqa


BooleanField.false_values = BooleanField.false_values + (False,)
